/*
 * eset_rtp (ESET Real-time file system protection module)
 * Copyright (C) 1992-2021 ESET, spol. s r.o.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * In case of any questions, you can contact us at ESET, spol. s r.o., Einsteinova 24, 851 01 Bratislava, Slovakia.
 */

#ifndef _ERTP_H
#define _ERTP_H

#include <linux/delay.h>
#include <linux/device.h>
#include <linux/cred.h>
#include <linux/err.h>
#include <linux/file.h>
#include <linux/fdtable.h>
#include <linux/fs.h>
#include <linux/kallsyms.h>
#include <linux/kdev_t.h>
#include <linux/kprobes.h>
#include <linux/mm_types.h>
#include <linux/module.h>
#include <linux/mount.h>
#include <linux/namei.h>
#include <linux/poll.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/radix-tree.h>
#include <linux/rcupdate.h>
#include <linux/version.h>
#include <linux/uaccess.h>
#include <linux/unistd.h>

#include "eset_rtp.h"

/*---------------------- HANDLERS - ertp_handlers.c ---------------------------*/
int ertp_handlers_init(uintptr_t k32, uintptr_t k64);
void ertp_handlers_deinit(void);

/*---------------------- HOOKS - ertp_hooks.c ---------------------------------*/

/* register hooks on open, close syscalls */
int ertp_hooks_init(uintptr_t t32, uintptr_t t64);

/* unregister eset hooks from syscalls */
void ertp_hooks_exit(void);

/* possible return values from antivirus filter */
enum ertp_flt_rv {
	ERTP_CONTINUE,
	ERTP_DENIED
};

/**
 * @brief call to antivirus filter to scan open file
 * @param struct file - open file
 * @return ERTP_CONTINUE when file is ok, ERTP_DENIED when operation was denied by filer
 */
enum ertp_flt_rv ertp_check_open(const struct file *file);

/**
 * @brief call to antivirus filter to scan closed file
 * @param struct file - closed file
 * @return ERTP_CONTINUE when file is ok, ERTP_DENIED when operation was denied by filer
 */
enum ertp_flt_rv ertp_check_close(const struct file *file);

/**
 * @brief call to antivirus filter to scan executed file
 * @param struct file - executed file
 * @return ERTP_CONTINUE when file is ok, ERTP_DENIED when operation was denied by filer
 */
enum ertp_flt_rv ertp_check_exec(const struct file *file);

/*------------------------ CONTROL DEVICE - ertp_dev.c ------------------------*/

/* create control device /dev/eset_hook */
int ertp_dev_init(void);

/* unregister and destroy /dev/eset_hook */
void ertp_dev_exit(void);

/*------------------------ EVENTS & REQUESTS - ertp_event.c -------------------*/

/* max timeout for scanner response in ms */
#define ERTP_MAX_TIMEOUT 10000

/* struct ertp_event - holds info for scanned event */
struct ertp_event {
	struct list_head req_list;            /* entry to request list */
	struct list_head scan_list;           /* entry to scanner list */
	struct list_head cache_waiting_list;  /* entry to cache, which the event is waiting for to be updated */
	struct completion wait;               /* wait for completion handler */
	atomic_t count;                       /* refcount */

	#ifndef f_dentry
		#define f_dentry f_path.dentry /* needed for kernel backward compatibility */
	#endif
	#define f_path_dentry f_path.dentry
	struct path f_path;               /* holding file path */
	unsigned int flags;               /* file->f_flags */
	struct file *file;                /* file for scan */

	int was_removed_from_req_list;    /* 1 if event was remove from request list, 0 otherwise */

	/* user copyable fields from type till end of struct*/
	int type;                          /* event type: ERTP_OPEN, ERTP_CLOSE, ERTP_EXEC or ERTP_REMOVE */
	int id;                            /* unique id number */
	int result;                        /* result from scanner; ERTP_ALLOW or ERTP_DENY */
	uint32_t scan_mask;               /* bit mask of applied scan settings */
	pid_t tgid;                       /* event's pid */
	int fd;                           /* file descriptor */
	unsigned long fs_magic;           /* file system magic number from include/linux/magic.h */
	uint64_t ino;                     /* inode number */
	uint64_t dev;                     /* device number */
	loff_t file_size;                 /* file size in bytes */
	time64_t ctime;                   /* tv_sec value of change time */
	time64_t mtime;                   /* tv_sec value of modified time */
	uint16_t mode;                    /* kstat mode */
};

/* increase/decrease refcount of @event */
struct ertp_event *ertp_event_get(struct ertp_event *event);
void ertp_event_put(struct ertp_event *event);

/* pop request from queue */
struct ertp_event *ertp_pop_request(void);

/* process scan request for @event_type operation for @file */
int ertp_process_request(const struct file *file, int event_type);

/* send notification message about removing file with @path and remove it from our internal cache*/
void ertp_remove_notify(struct path *path);

/* mark event as done */
void ertp_event_done(struct ertp_event *event);

/* update cache with the current event results */
void ertp_update_cache(const struct ertp_event *event);

/* open/close file in @event - needed to send it to scanner */
int ertp_open_event_file(struct ertp_event *event);
void ertp_close_event_file(struct ertp_event *event);

/* install file descriptor for @event into user space */
void ertp_install_fd(struct ertp_event *event);

/* copy event data to @buf */
ssize_t ertp_event_to_buf(char __user *buf, size_t size,
		const struct ertp_event *event);

/* assign @event to be handled by current process */
int ertp_add_reply(struct ertp_event *event);
int ertp_request_queue_empty(void);

/* enable/disable sending event to antivirus filter */
void ertp_start_accept(void);
void ertp_stop_accept(void);
/* test whether any antivirus is registred to module */
int ertp_is_stopped(void);

/* remove all request from cache */
void ertp_rem_requests(void);

/* get ertp_event for user space @response */
struct ertp_event *ertp_get_reply(struct eset_rtp_response *response);

/**
 * ertp_get_full_path - get full file path of @event
 *
 * use buffer @buf with size @size to store the path
 * return pointer to path string
 */
static inline const char *ertp_get_full_path(const struct ertp_event *event, char *buf, size_t size) {
	return d_path(&event->f_path, buf, size);
}

/* create/destroy event cache */
int ertp_events_init(void);
void ertp_events_exit(void);

extern wait_queue_head_t ertp_request_available;

/*------------------------- PATHS - ertp_path.c -------------------------------*/

struct ertp_path {
	struct list_head list; /* list of paths */
	char *path_name;       /* null-terminated path */
	size_t path_len;       /* len of path */
	atomic_t count;        /* refcount */
};

/* alloc new ertp_path */
struct ertp_path *ertp_path_alloc(const char* path_name);

/* increase/decrease refcount of @path */
struct ertp_path *ertp_path_get(struct ertp_path* path);
void ertp_path_put(struct ertp_path* path);

/*----------------------- EXCLUDES - ertp_excludes.c --------------------------*/

/* return true if current process is excluded, return false if it is not or in case of error*/
bool ertp_proc_excluded(void);
/* return true if @file is excluded, return false if it is not or in case of error*/
bool ertp_file_excluded(const struct file *file);

/* Show or store excluded processes. Called from sysfs. */
int ertp_excluded_proc_show(char *buf, int size);
int ertp_excluded_proc_store(const char *buf, int size);

/* Show or store excluded files. Called from sysfs. */
int ertp_excluded_files_show(char *buf, int size);
int ertp_excluded_files_store(const char *buf, int size);

/* clean cache of excludes*/
void ertp_excluded_proc_clear(void);
void ertp_excluded_files_clear(void);
void ertp_excludes_clear(void);

/*------------------------ SCANNER - escanner.c ---------------------------*/

/* struct ertp_scanner - information about scanner connected to eset_rtp */
/* Note: current implementation supports only one connected scanner */
struct ertp_scanner {
	struct list_head events; /* list of scanned events */
	spinlock_t lock;         /* spin lock */
	atomic_t   tgid;         /* process tgit */
};

extern struct ertp_scanner escanner;

/* return true if there is no registered scanner */
static inline int ertp_no_scanner(void)
{
	return atomic_read(&escanner.tgid) == -1;
}

/* return true if current process is registered scanner */
static inline int ertp_is_scanner(void)
{
	return atomic_read(&escanner.tgid) == current->tgid;
}

/* register process with @tgid as current scanner */
/* there has to be no registered scanner or return -EEXIST */
int ertp_scanner_reg(pid_t tgid);

/* remove registered scanner and clean all its events */
void ertp_scanner_rem(void);

/* add/remove event to/from scanner */
int ertp_scan_add_event(struct ertp_event *event);
void ertp_scan_rem_event(struct ertp_event *event);

/* pop event with @id from scanner list */
struct ertp_event *ertp_scan_pop_event(int id);

/*--------------------- /sys/module control - ertp_sysfs.c ------------------*/

/* init /sys/module/eset_rtp control directory */
int ertp_sysfs_init(void);
/* delete sysfs directories */
void ertp_sysfs_exit(void);

/*--------------- cache of scanned result - ertp_cache.c --------------------*/

/* cache stores just clean results, infected files are sent to scanner on every event
   (there is just the optimization with no parallel scans of the same file) */
struct ertp_inode_data {
	struct list_head    time_list;  /* entry to inode data list sorted according to last usage time */
	const struct inode *inode;      /* ID */
	unsigned long       i_ino;      /* inode number */
	dev_t               s_dev;      /* internal device number */
	int                 result;     /* last result from scanner: ERTP_ALLOW or ERTP_UNKNOWN - only clean results are cached */
	uint32_t            scan_mask;  /* bit mask of applied scan settings for this record */

#if LINUX_VERSION_CODE < KERNEL_VERSION(4,18,0)
	struct timespec   cached_ctime; /* last change time of cached data */
	struct timespec   running_scan_ctime; /* last change time of currently scanned file */
#else
	struct timespec64 cached_ctime; /* last change time of cached data */
	struct timespec64 running_scan_ctime; /* last change time of currently scanned file */
#endif

	atomic_t           count; /* refcount */
	spinlock_t         lock; /* lock protecting data */
	struct ertp_event *scan_in_progress; /* non-nullpointer if this file is currently scanned and eset_rtp is waiting for response */
	struct list_head   waiting_events; /* entry to list of events waiting for this result */
};

/* allocate new cache data for event */
struct ertp_inode_data *ertp_data_alloc(const struct ertp_event *event);

/* fill event specific fields to @data (before scan) */
void ertp_fill_data_unlocked(struct ertp_inode_data *data, const struct ertp_event *event);
static inline void ertp_fill_data(struct ertp_inode_data *data, const struct ertp_event *event) {
	spin_lock(&data->lock);
	ertp_fill_data_unlocked(data, event);
	spin_unlock(&data->lock);
}

/* fill event specific fields to @data (after scan) */
void ertp_update_data(struct ertp_inode_data *data, const struct ertp_event *event);

/* get/release reference of @data */
void ertp_data_get(struct ertp_inode_data *data);
void ertp_data_put(struct ertp_inode_data *data);

/* insert a @data in the cache */
int ertp_insert_data_to_cache(struct ertp_inode_data *data);

/* look up inode data in cache */
struct ertp_inode_data *ertp_get_data(const struct inode *inode);

/* remove @inode from cache and release a reference */
void ertp_remove_inode_from_cache(const struct inode *inode);

/* remove all data from cache */
void ertp_cache_clear(void);

/* change max cache size and flush it */
void ertp_set_cache_size(size_t max_size);

/* init cache of scanned files */
int ertp_cache_init(void);
/* clear and releas cache resources */
void ertp_cache_exit(void);

/* add event to be waiting for new update in cache - used when the same file is already being scanned */
/* return true of success */
bool ertp_add_to_cache_for_waiting_on_update(struct ertp_event *event);

/* attach ertp_event to cache data and set scan is running */
void ertp_cache_set_scan_is_running(struct ertp_inode_data *data, struct ertp_event *scan_in_progress);
/* release waiting events and set scan is finished and update date by scan @result and @scan_mask */
void ertp_cache_set_scan_is_finished(struct ertp_inode_data *data, int result, uint32_t scan_mask);

/* get settings mask of event @type */
uint32_t ertp_get_scan_mask_of_event_type(int type);

/* set settings @mask of event @type */
void ertp_set_scan_mask_of_event_type(uint32_t mask, int type);

/*------------------------ utilities - ertp_utils.c -------------------------*/

/* lookup address of symbol, return 0 on error */
unsigned long ertp_lookup_address(const char *symbol_name);

static inline struct file *ertp_dentry_open(const struct path *path, int flags) {
#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,6,0))
	return dentry_open(dget(path->dentry), mntget(path->mnt), flags,
			current_cred());
#else
	return dentry_open(path, flags, current_cred());
#endif
}

static inline int ertp_vfs_getattr(const struct path *path, struct kstat *attr_info) {
#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,9,0))
	return vfs_getattr(path->mnt, path->dentry, attr_info);
#elif (LINUX_VERSION_CODE < KERNEL_VERSION(4,11,0))
	return vfs_getattr((struct path *)path, attr_info);
#else
	return vfs_getattr(path, attr_info, 0, 0);
#endif
}

#ifdef ERTP_DEBUG
/* get/release filename from userspace to kernel space */
/* Not very optimal, use only for debug purposes */
char *ertp_getname(const char __user * filename);
static inline void ertp_putname(char * filename) { if (!IS_ERR_OR_NULL(filename)) __putname(filename); }
#	define ERPT_UNKONWN "[unknown]"
#endif

/*----------------------------- LOGS ----------------------------------------*/

#define ertp_pr_error(fmt, ...) \
	printk(KERN_ERR ESET_RTP "(%s): " pr_fmt(fmt) "\n", __PRETTY_FUNCTION__, ##__VA_ARGS__)
#define ertp_pr_warning(fmt, ...) \
	printk(KERN_WARNING ESET_RTP "(%s): " pr_fmt(fmt) "\n", __PRETTY_FUNCTION__, ##__VA_ARGS__)
#define ertp_pr_info(fmt, ...) \
	printk(KERN_INFO ESET_RTP ": " pr_fmt(fmt) "\n", ##__VA_ARGS__)

/*----------------------------- DEBUG ---------------------------------------*/

#ifdef ERTP_DEBUG
#	define DBG_BUG_ON(cond) BUG_ON(cond)
#	define ertp_pr_debug(fmt, ...) \
		printk(KERN_DEBUG ESET_RTP ": %s:%d:%s:" pr_fmt(fmt) "\n", __FILE__, __LINE__, __PRETTY_FUNCTION__, ##__VA_ARGS__)
#else
#	define DBG_BUG_ON(cond)
#	define ertp_pr_debug(fmt, ...)
#endif

#endif /* _ERTP_H */
