/*
 * eset_rtp (ESET Real-time file system protection module)
 * Copyright (C) 1992-2021 ESET, spol. s r.o.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * In case of any questions, you can contact us at ESET, spol. s r.o., Einsteinova 24, 851 01 Bratislava, Slovakia.
 */

#include "ertp.h"

/* memory cache where data are allocated */
static struct kmem_cache *ertp_scan_cache = NULL;

/* radix tree for quick look up */
static RADIX_TREE(ertp_scan_radix_tree, GFP_ATOMIC);

/* spinlock protecting radix tree */
static DEFINE_SPINLOCK(ertp_scan_radix_tree_lock);

static uint32_t ertp_open_scan_mask = 0;
static uint32_t ertp_close_scan_mask = 0;
static uint32_t ertp_exec_scan_mask = 0;

/* spinlock protecting scan masks */
static DEFINE_SPINLOCK(ertp_scan_masks_lock);

uint32_t ertp_get_scan_mask_of_event_type(int type)
{
	uint32_t mask;

	spin_lock(&ertp_scan_masks_lock);
	switch (type) {
		case ERTP_OPEN:
			mask = ertp_open_scan_mask;
			break;
		case ERTP_CLOSE:
			mask = ertp_close_scan_mask;
			break;
		case ERTP_EXEC:
			mask = ertp_exec_scan_mask;
			break;
		default:
			ertp_pr_error("Scan mask for type %d does not exist", type);
			mask = ertp_open_scan_mask;
	};
	spin_unlock(&ertp_scan_masks_lock);

	return mask;
}

void ertp_set_scan_mask_of_event_type(uint32_t mask, int type)
{
	spin_lock(&ertp_scan_masks_lock);
	switch (type) {
		case ERTP_OPEN:
			ertp_open_scan_mask = mask;
			break;
		case ERTP_CLOSE:
			ertp_close_scan_mask = mask;
			break;
		case ERTP_EXEC:
			ertp_exec_scan_mask = mask;
			break;
		default:
			ertp_pr_error("Scan mask for type %d does not exist", type);
	};
	spin_unlock(&ertp_scan_masks_lock);
}

struct ertp_scan_time_list_t {
	struct list_head root;
	spinlock_t lock;
	size_t size;
	size_t max_size;
};

/* list of inode data, sorted according last access time */
static struct ertp_scan_time_list_t ertp_scan_time_list = {
	.root = LIST_HEAD_INIT(ertp_scan_time_list.root),
	.lock = __SPIN_LOCK_UNLOCKED(ertp_scan_time_list.lock),
	.size = 0,
	.max_size = 100000,
};

void ertp_fill_data_unlocked(struct ertp_inode_data *data, const struct ertp_event *event)
{
	const struct inode *inode = event->f_path_dentry->d_inode;

	data->i_ino = inode->i_ino;
	data->s_dev = inode->i_sb->s_dev;
	data->result = ERTP_UNKNOWN;
	data->cached_ctime = inode->i_ctime;
	data->scan_mask = event->scan_mask;
}

bool ertp_add_to_cache_for_waiting_on_update(struct ertp_event *event)
{
	bool ret = false;
	struct ertp_inode_data *data = ertp_get_data(event->f_path_dentry->d_inode);

	if (!data) {
		ertp_pr_debug("cache record for inode: %lu not found", data->i_ino);
		return false;
	}

	spin_lock(&data->lock);
	if (data->scan_in_progress) {
		ertp_event_get(event);
		list_add_tail(&event->cache_waiting_list, &data->waiting_events);
		ret = true;
		ertp_pr_debug("event for inode: %lu added to cache data and waiting for update", data->i_ino);
	}
	spin_unlock(&data->lock);

	ertp_data_put(data);
	return ret;
}

void ertp_cache_set_scan_is_running(struct ertp_inode_data *data, struct ertp_event *scan_in_progress)
{
	if (!data || !scan_in_progress)
		return;

	spin_lock(&data->lock);
	if (!data->scan_in_progress) {
		data->running_scan_ctime = scan_in_progress->f_path_dentry->d_inode->i_ctime;
		data->scan_in_progress = ertp_event_get(scan_in_progress);
	}
	spin_unlock(&data->lock);
}

void ertp_cache_set_scan_is_finished(struct ertp_inode_data *data, int result, uint32_t scan_mask)
{
	struct ertp_event *attached_event;
	struct ertp_event *event;
	struct ertp_event *tmp;
	LIST_HEAD(finished_events_list);

	if (!data)
		return;

	spin_lock(&data->lock);
	attached_event = data->scan_in_progress;
	data->scan_in_progress = NULL;
	/* update just result (caching only clean files without any flags) and scan mask fields; ctime, inode etc. stay from the moment of scan beginning */
	data->result = (result & ERTP_DO_NOT_CACHE || result & ERTP_DENY) ? ERTP_UNKNOWN : result;
	data->scan_mask = scan_mask;

	list_for_each_entry_safe(event, tmp, &data->waiting_events, cache_waiting_list) {
		list_move_tail(&event->cache_waiting_list, &finished_events_list);
	}
	spin_unlock(&data->lock);

	ertp_event_put(attached_event);
	list_for_each_entry_safe(event, tmp, &finished_events_list, cache_waiting_list) {
		list_del_init(&event->cache_waiting_list);
		event->result = result; // use real result, e.g. ERTP_DENY for infected files
		ertp_event_done(event);
		ertp_pr_debug("event for inode: %lu unblocked via cache update", data->i_ino);
		ertp_event_put(event);
	}
}

void ertp_update_data(struct ertp_inode_data *data, const struct ertp_event *event)
{
	ertp_cache_set_scan_is_finished(data, event->result, event->scan_mask);

	spin_lock(&ertp_scan_time_list.lock);
	if (!list_empty(&data->time_list)) {
		/* move data at the end of time cache list */
		list_move_tail(&data->time_list, &ertp_scan_time_list.root);
	}
	spin_unlock(&ertp_scan_time_list.lock);
}

struct ertp_inode_data *ertp_data_alloc(const struct ertp_event *event)
{
	struct ertp_inode_data *data;
	const struct inode *inode = event->f_path_dentry->d_inode;

	if (!inode)
		return ERR_PTR(-EINVAL);

	data = kmem_cache_zalloc(ertp_scan_cache, GFP_KERNEL);
	if (!data)
		return ERR_PTR(-ENOMEM);

	INIT_LIST_HEAD(&data->time_list);
	data->inode = inode;
	ertp_fill_data_unlocked(data, event);
	atomic_set(&data->count, 1);
	spin_lock_init(&data->lock);
	INIT_LIST_HEAD(&data->waiting_events);
	data->scan_in_progress = NULL;

	return data;
}

void ertp_data_get(struct ertp_inode_data *data)
{
	atomic_inc(&data->count);
}

void ertp_data_put(struct ertp_inode_data *data)
{
	if (!data)
		return;

	if (atomic_dec_and_test(&data->count)) {
		ertp_cache_set_scan_is_finished(data, ERTP_UNKNOWN, 0);
		kmem_cache_free(ertp_scan_cache, data);
	}
}

static void ertp_remove_oldest_10_percent_of_cache(void)
{
	struct ertp_inode_data *data;
	struct ertp_inode_data *tmp;
	LIST_HEAD(removed_data_list);
	size_t i = 0;

	// move 10% of entries from time list to a temporary one
	spin_lock(&ertp_scan_time_list.lock);
	for (; i < ertp_scan_time_list.size / 10; ++i) {
		struct ertp_inode_data *oldest_data = container_of(ertp_scan_time_list.root.next,
			struct ertp_inode_data, time_list);
		list_move_tail(&oldest_data->time_list, &removed_data_list);
	}

	ertp_pr_debug("Max cache size %zu reached removing %zu oldest records", ertp_scan_time_list.max_size, i);
	ertp_scan_time_list.size -= i;
	spin_unlock(&ertp_scan_time_list.lock);

	// remove them from radix tree
	spin_lock(&ertp_scan_radix_tree_lock);
	list_for_each_entry(data, &removed_data_list, time_list) {
		tmp = radix_tree_delete(&ertp_scan_radix_tree, (long)data->inode);
		if (tmp) {
			ertp_data_put(tmp); // paired with ertp_data_get() in ertp_insert_data_to_cache()
		}
	}
	spin_unlock(&ertp_scan_radix_tree_lock);

	// release all (no longer used) entries
	list_for_each_entry_safe(data, tmp, &removed_data_list, time_list) {
		ertp_pr_debug("deleting id: %d, on addr: %p, with refcount: %d", i++, data, atomic_read(&data->count));
		list_del_init(&data->time_list);
		ertp_data_put(data); // paired with ertp_data_get() in ertp_insert_data_to_time_list()
	}
}

static void ertp_insert_data_to_time_list(struct ertp_inode_data *data)
{
	bool max_size_exceeded;
	ertp_data_get(data);
	spin_lock(&ertp_scan_time_list.lock);
	list_add_tail(&data->time_list, &ertp_scan_time_list.root);
	++ertp_scan_time_list.size;
	max_size_exceeded = ertp_scan_time_list.size > ertp_scan_time_list.max_size;
	spin_unlock(&ertp_scan_time_list.lock);

	if (max_size_exceeded) {
		ertp_remove_oldest_10_percent_of_cache();
	}
}

static void ertp_remove_data_from_time_list(struct ertp_inode_data *data)
{
	bool removed = false;
	spin_lock(&ertp_scan_time_list.lock);
	if (!list_empty(&data->time_list)) {
		list_del_init(&data->time_list);
		--ertp_scan_time_list.size;
		removed = true;
	}
	spin_unlock(&ertp_scan_time_list.lock);

	if (removed) {
		ertp_data_put(data); // paired with ertp_data_get() in ertp_insert_data_to_time_list()
	}
}

int ertp_insert_data_to_cache(struct ertp_inode_data *data)
{
	int err = radix_tree_preload(GFP_KERNEL);
	if (err)
		return err; // -ENOMEM

	ertp_data_get(data);
	spin_lock(&ertp_scan_radix_tree_lock);
	{
		err = radix_tree_insert(&ertp_scan_radix_tree,
			(long)data->inode, data);
	}
	spin_unlock(&ertp_scan_radix_tree_lock);

	if (err) {
		// e.g. -EEXIST
		ertp_data_put(data);
		ertp_pr_debug("Cannot insert data to cache, inode %lu, device number %u, error %d",
			data->inode->i_ino, new_encode_dev(data->inode->i_sb->s_dev), err);
	}

	radix_tree_preload_end();

	if (!err)
		ertp_insert_data_to_time_list(data);

	return err;
}

struct ertp_inode_data *ertp_get_data(const struct inode *inode)
{
	struct ertp_inode_data *data;

	if (!inode)
		return NULL;

	spin_lock(&ertp_scan_radix_tree_lock);
	data = radix_tree_lookup(&ertp_scan_radix_tree, (long)inode);
	if (data)
		ertp_data_get(data);
	spin_unlock(&ertp_scan_radix_tree_lock);

	return data;
}

/* remove @inode from cache and release a reference */
void ertp_remove_inode_from_cache(const struct inode *inode)
{
	struct ertp_inode_data *data;
	if (unlikely(!inode)) {
		ertp_pr_debug("invalid inode!");
		return;
	}

	spin_lock(&ertp_scan_radix_tree_lock);
	data = radix_tree_delete(&ertp_scan_radix_tree, (long)inode);
	spin_unlock(&ertp_scan_radix_tree_lock);

	if (data) {
		ertp_pr_debug("data: %p removed from cache, refcount: %d", data, atomic_read(&data->count));
		// call (the last) ertp_data_put outside lock
		ertp_remove_data_from_time_list(data);
		ertp_data_put(data); // paired with ertp_data_get() in ertp_insert_data_to_cache()
	}
}

void ertp_cache_clear(void)
{
	struct ertp_inode_data *data;
	struct ertp_inode_data *tmp;
	LIST_HEAD(removed_data_list);
#ifdef ERTP_DEBUG
	int i = 0;
#endif
	// move all entries from time list to a temporary one
	spin_lock(&ertp_scan_time_list.lock);
	list_for_each_entry_safe(data, tmp, &ertp_scan_time_list.root, time_list) {
		list_move_tail(&data->time_list, &removed_data_list);
	}
	ertp_scan_time_list.size = 0;
	spin_unlock(&ertp_scan_time_list.lock);

	// remove them from radix tree
	spin_lock(&ertp_scan_radix_tree_lock);
	list_for_each_entry(data, &removed_data_list, time_list) {
		tmp = radix_tree_delete(&ertp_scan_radix_tree, (long)data->inode);
		if (tmp) {
			ertp_data_put(tmp); // paired with ertp_data_get() in ertp_insert_data_to_cache()
		}
	}
	spin_unlock(&ertp_scan_radix_tree_lock);

	// release all (no longer used) entries
	list_for_each_entry_safe(data, tmp, &removed_data_list, time_list) {
		ertp_pr_debug("deleting id: %d, on addr: %p, with refcount: %d", i++, data, atomic_read(&data->count));
		list_del_init(&data->time_list);
		ertp_data_put(data); // paired with ertp_data_get() in ertp_insert_data_to_time_list()
	}
}

void ertp_set_cache_size(size_t max_size)
{
	// clear old cached data
	ertp_cache_clear();
	spin_lock(&ertp_scan_time_list.lock);
	ertp_scan_time_list.max_size = max_size;
	spin_unlock(&ertp_scan_time_list.lock);
}

int ertp_cache_init(void)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
	ertp_scan_cache = kmem_cache_create("ertp_scan_cache",
		sizeof(struct ertp_inode_data),
		0, SLAB_RECLAIM_ACCOUNT, NULL, NULL);
#else
	ertp_scan_cache = kmem_cache_create("ertp_scan_cache",
		sizeof(struct ertp_inode_data),
		0, SLAB_RECLAIM_ACCOUNT, NULL);
#endif

	if (!ertp_scan_cache)
		return -ENOMEM;

	return 0;
}

void ertp_cache_exit(void)
{
	ertp_cache_clear();
	kmem_cache_destroy(ertp_scan_cache);
}
