/*
 * eset_rtp (ESET Real-time file system protection module)
 * Copyright (C) 1992-2021 ESET, spol. s r.o.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * In case of any questions, you can contact us at ESET, spol. s r.o., Einsteinova 24, 851 01 Bratislava, Slovakia.
 */

#include "ertp.h"

static int ertp_dev_open(struct inode *inode, struct file *file)
{
	int ret;

	/* must be open with O_WRITE */
	if (!(file->f_mode & FMODE_WRITE))
		return -EINVAL;

	ret = ertp_scanner_reg(current->tgid);
	if (ret)
		return ret;

	ertp_start_accept();
	return 0;
}

static int ertp_dev_release(struct inode *inode, struct file *file)
{
	if (ertp_is_scanner()) {
		ertp_scanner_rem();
		ertp_stop_accept();
		ertp_rem_requests();
	}

	return 0;
}

///@brief pop, prepare and insert to buffer next `erpt_event` from queue
///@return length of event inserted to buffer @buf, zero if no more available, -1 if error occured
static ssize_t ertp_pop_and_insert_request_to_buffer(char __user *buf, size_t size)
{
	ssize_t len;
	struct ertp_event *event;

	// not enough space for next event in buffer
	if (size < sizeof(struct eset_rtp_event))
		return 0;

	// no more request
	event = ertp_pop_request();
	if (!event)
		return 0;

	/* for remove notify these informations are not needed */
	if (event->type != ERTP_REMOVE) {
		if (ertp_open_event_file(event) || ertp_scan_add_event(event))
			goto error;

		ertp_install_fd(event);
	}

	len = ertp_event_to_buf(buf, size, event);
	if (len < 0)
		goto error;

	ertp_pr_debug("event id %d sent to user space", event->id);
	ertp_event_put(event);
	return len;
error:
	ertp_close_event_file(event);
	event->result = ERTP_ALLOW;
	ertp_event_done(event);
	ertp_event_put(event);
	return -1;
}

static ssize_t ertp_dev_read(struct file *file, char __user *buf,
		size_t size, loff_t *pos)
{
	ssize_t len = 0;
#ifdef ERTP_DEBUG
	int i = 0; // debug counter
#endif

	// not enough space for a single event in buffer
	if (size < sizeof(struct eset_rtp_event))
		return -EINVAL;
	while (true) {
		ssize_t ret = ertp_pop_and_insert_request_to_buffer(buf + len, size - len);

		if (ret <= 0)
			break;

		len += ret;
#ifdef ERTP_DEBUG
		i++; // debug counter
#endif
	}

	ertp_pr_debug("%d events inserted into buffer of one read() call", i);
	return len;
}

static ssize_t ertp_dev_write(struct file *file, const char __user *buf,
		size_t size, loff_t *pos)
{
	struct eset_rtp_response response;
	size_t offset = 0;

	if (size < sizeof(response)) {
		ertp_pr_debug("invalid length(%zu) of response", size);
		return -EINVAL;
	}

	/* check if there is enough space in buf for next response */
	while (size - offset >= sizeof(response)) {
		struct ertp_event *event;
		if (copy_from_user(&response, buf + offset, sizeof(response))) {
			ertp_pr_debug("copy_from_user(from = %p) failed", buf + offset);
			return -EFAULT;
		}

		if (response.size != sizeof(response)) {
			ertp_pr_debug("invalid message size member(%u), this version\
				of kernel module requires size: %zu", response.size, sizeof(response));
			return -EINVAL;
		}

		event = ertp_get_reply(&response);
		if (IS_ERR(event))
			return PTR_ERR(event);

		ertp_event_done(event);
		ertp_update_cache(event);
		ertp_event_put(event);

		offset += response.size;
	}

	return offset;
}

static unsigned int ertp_poll(struct file *file, poll_table *wait)
{
	unsigned int mask;

	poll_wait(file, &ertp_request_available, wait);

	mask = POLLOUT | POLLWRNORM;

	if (!ertp_request_queue_empty())
		mask |= POLLIN | POLLRDNORM;

	return mask;
}

static struct class  *ertp_class;
static struct device *ertp_device;
static dev_t          ertp_dev;

static struct file_operations ertp_fops = {
	.owner =   THIS_MODULE,
	.open =    ertp_dev_open,
	.release = ertp_dev_release,
	.read =    ertp_dev_read,
	.write =   ertp_dev_write,
	.poll =    ertp_poll
};

int ertp_dev_init(void)
{
	int major = register_chrdev(0, ESET_RTP, &ertp_fops);
	if (major < 0)
		return major;

	ertp_dev = MKDEV(major, 0);

	ertp_class = class_create(THIS_MODULE, ESET_RTP);
	if (IS_ERR(ertp_class)) {
		unregister_chrdev(major, ESET_RTP);
		return PTR_ERR(ertp_class);
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,27)
	ertp_device = device_create(ertp_class, NULL, ertp_dev, ESET_RTP);
#else
	ertp_device = device_create(ertp_class, NULL, ertp_dev, NULL, ESET_RTP);
#endif
	if (IS_ERR(ertp_device)) {
		class_destroy(ertp_class);
		unregister_chrdev(major, ESET_RTP);
		return PTR_ERR(ertp_device);
	}

	return 0;
}

void ertp_dev_exit(void)
{
	device_destroy(ertp_class, ertp_dev);
	class_destroy(ertp_class);
	unregister_chrdev(MAJOR(ertp_dev), ESET_RTP);

	ertp_dev_release(NULL, NULL); // remove pending requests if any remain
}
