/*
 * eset_rtp (ESET Real-time file system protection module)
 * Copyright (C) 1992-2021 ESET, spol. s r.o.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * In case of any questions, you can contact us at ESET, spol. s r.o., Einsteinova 24, 851 01 Bratislava, Slovakia.
 */

#include "ertp.h"

#define SYS_CALL_TABLE "sys_call_table"
#define SYS_CALL_TABLE_32 "ia32_sys_call_table"

static inline bool ertp_file_empty(const struct path *path)
{
	struct kstat attr_info;
	if (ertp_vfs_getattr(path, &attr_info) < 0) {
		return false;
	}

	return attr_info.size == 0;
}

static int ertp_should_check(const struct file *file, int type)
{
	if (ertp_is_stopped())
		return 0;

	/* scanner process */
	if (ertp_is_scanner())
		return 0;

	/* current process is excluded */
	if (ertp_proc_excluded())
		return 0;

	/* inode doesn't exist */
	if (!file->f_dentry->d_inode)
		return 0;

	/* not a regular file */
	if(!S_ISREG(file->f_dentry->d_inode->i_mode))
		return 0;

	/* empty file */
	if (ertp_file_empty(&file->f_path))
		return 0;

	/* closing O_RDONLY file */
	if (type == ERTP_CLOSE && ((file->f_flags & O_ACCMODE) == O_RDONLY))
		return 0;

	/* file is excluded */
	if (ertp_file_excluded(file))
		return 0;

	return 1;
}

static inline enum ertp_flt_rv ertp_eval_res(int rv)
{
	switch(ERTP_WITHOUT_FLAGS(rv)) {
		case ERTP_DENY:
			return ERTP_DENIED;
		default:
			return ERTP_CONTINUE;
	}
}

static inline bool ertp_scan_mask_usable(uint32_t scan_mask_of_running_scan, int type_of_new_event)
{
	uint32_t wanted_mask = ertp_get_scan_mask_of_event_type(type_of_new_event);
	return (scan_mask_of_running_scan & wanted_mask) == wanted_mask;
}

/* return true if event which is scanned can be used */
static inline bool ertp_cache_same_scan_in_progress(const struct inode *inode, const struct ertp_inode_data *data, int type)
{
	if (data->scan_in_progress && (data->scan_in_progress->ino == inode->i_ino)
	    && (data->scan_in_progress->dev == new_encode_dev(inode->i_sb->s_dev))
	    && (data->running_scan_ctime.tv_sec == inode->i_ctime.tv_sec)
	    && (data->running_scan_ctime.tv_nsec == inode->i_ctime.tv_nsec)
	    && ertp_scan_mask_usable(data->scan_in_progress->scan_mask, type)) {
		return true;
	}

	return false;
}

/*
 @return  ERTP_UNKNOWN - result can not be used (file dismatch or only open event cached)
          ERTP_MODIFIED - file has changed
          ERTP_SCAN_IN_PROGRESS - scan of the same file is already running
          ERTP_ALLOW or ERTP_DENY - when cached result can be used
*/
static inline int ertp_get_cache_result(const struct inode *inode, const struct ertp_inode_data *data, int type)
{
	/* inode mismatched - this is not the cached file! */
	if ((data->i_ino != inode->i_ino) || (data->s_dev != inode->i_sb->s_dev)) {
		return ERTP_UNKNOWN;
	}

	/* result of currently scanned file can be used */
	if (ertp_cache_same_scan_in_progress(inode, data, type)) {
		return ERTP_SCAN_IN_PROGRESS;
	}

	/* file is changed */
	if ((data->cached_ctime.tv_sec != inode->i_ctime.tv_sec)
			|| (data->cached_ctime.tv_nsec != inode->i_ctime.tv_nsec)) {
		return ERTP_MODIFIED;
	}

	/* cached result has weaker scan settings mask */
	if (!ertp_scan_mask_usable(data->scan_mask, type))
		return ERTP_UNKNOWN;

	return data->result;
}

static int ertp_check_cache(const struct file *file, int type)
{
	int result = ERTP_UNKNOWN;
	const struct inode *inode = file->f_dentry->d_inode;
	struct ertp_inode_data *data = ertp_get_data(inode);

	if (!data)
		return ERTP_UNKNOWN;

	spin_lock(&data->lock);
	result = ertp_get_cache_result(inode, data, type);
	spin_unlock(&data->lock);

	ertp_data_put(data);

#if 0
	{
		static char BUF[256];
		const char *file_name = simple_dname(file->f_path_dentry, BUF, sizeof(BUF));
		if (!IS_ERR(file_name)) {
			if ((result == ERTP_ALLOW) || (result == ERTP_DENY)) {
				ertp_pr_debug("using cached result: %s for event: %s, file: %s",
					(result == ERTP_ALLOW) ? "Allow" : "Deny",
					(type == ERTP_OPEN) ? "Open" : "Close",
					file_name
				);
			} else {
				ertp_pr_debug("cache result (%d) -> need to scan event: %s, file: %s",
					result,
					(type == ERTP_OPEN) ? "Open" : "Close",
					file_name
				);
			}
		}
	}
#endif
	return result;
}

static enum ertp_flt_rv ertp_check_file(const struct file *file, int type)
{
	int rv;

	if (!ertp_should_check(file, type)) {
		return ERTP_CONTINUE;
	}

	rv = ertp_check_cache(file, type);
	switch (rv) {
		case ERTP_ALLOW:
		case ERTP_DENY:
			return ertp_eval_res(rv);
		case ERTP_UNKNOWN:
			break;
		case ERTP_MODIFIED:
			type |= ERTP_MODIFIED; // add modified flag to request type
			break;
		case ERTP_SCAN_IN_PROGRESS:
			type |= ERTP_SCAN_IN_PROGRESS;
			break;
		default:
			ertp_pr_error("Not handled cache result %d", rv);
	}

	rv = ertp_process_request(file, type);
	return ertp_eval_res(rv);
}

enum ertp_flt_rv ertp_check_open(const struct file *file)
{
	return ertp_check_file(file, ERTP_OPEN);
}

enum ertp_flt_rv ertp_check_close(const struct file *file)
{
	return ertp_check_file(file, ERTP_CLOSE);
}

enum ertp_flt_rv ertp_check_exec(const struct file *file)
{
	return ertp_check_file(file, ERTP_EXEC);
}

int ertp_hooks_init(uintptr_t t32, uintptr_t t64)
{
	int rv;

/* try to find syscall tables dynamically, overriding provided values if found */
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,32))
	uintptr_t k32, k64 = ertp_lookup_address(SYS_CALL_TABLE);

	if (k64) {
		printk(KERN_DEBUG ESET_RTP ": found 64-bit syscall table at address %p\n", (void *)k64);
		t64 = k64;
	}

#ifdef CONFIG_IA32_EMULATION
	k32 = ertp_lookup_address(SYS_CALL_TABLE_32);

	if (k32) {
		printk(KERN_DEBUG ESET_RTP ": found 32-bit syscall table at address %p\n", (void *)k32);
		t32 = k32;
	}
#endif /* CONFIG_IA32_EMULATION */
#endif

	if (!t64) {
		ertp_pr_error("cannot find 64-bit syscall table");
		return -ENOENT;
	}

#ifdef CONFIG_IA32_EMULATION
	if (!t32) {
		ertp_pr_error("cannot find 32-bit syscall table");
		return -ENOENT;
	}
#endif /* CONFIG_IA32_EMULATION */

	if ((rv = ertp_handlers_init(t32, t64)) != 0) {
		return rv;
	}

	return 0;
}

void ertp_hooks_exit(void)
{
	ertp_handlers_deinit();
}
