/*
 * eset_rtp (ESET Real-time file system protection module)
 * Copyright (C) 1992-2021 ESET, spol. s r.o.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * In case of any questions, you can contact us at ESET, spol. s r.o., Einsteinova 24, 851 01 Bratislava, Slovakia.
 */

#ifndef _ESET_RTP_H
#define _ESET_RTP_H

#define ESET_RTP_FULL_NAME "ESET Real-time file system protection module"
#define ESET_RTP "eset_rtp" // external name
#define ERTP     "ertp"     // internal name

#define ERTP_VERSION "1.10"

#include <linux/types.h>

/*------------------Communication engine with user space---------------------*/

/* Path to a device to communicate with ertp
module */
#define ERTP_DEV "/dev/" ESET_RTP

#pragma pack(4)

#define ERTP_OPEN         1
#define ERTP_CLOSE        2
#define ERTP_EXEC         4
#define ERTP_BASIC_EVENTS_FLAG 0xF
#define ERTP_REMOVE       0x10 // only file_id and device_number are set and eset_rtp_response is not required

#define ERTP_UNKNOWN      0
#define ERTP_ALLOW        1
#define ERTP_DENY         2
#define ERTP_DO_NOT_CACHE      0x00010000 /* use when you don't want to cache result, e.g. because of excludes */
#define ERTP_MODIFIED          0x00020000 /* we had cached this file, but it has changed */
#define ERTP_SCAN_IN_PROGRESS  0x00040000 /* scan we need is already in progress, so let's wait for its result */
#define ERTP_FLAGS_MASK        0xFFFF0000

#define ERTP_WITHOUT_FLAGS(result) (result & ~ERTP_FLAGS_MASK) /* get pure ERTP_ALLOW or ERTP_ALLOW status */

struct eset_rtp_event {
	uint32_t size;       /* message size - for backward compatibility; more members can be added in future */
	int32_t  id;         /* unique id */
	int32_t  fd;         /* file descriptor of the file connected with event */
	int32_t  pid;        /* pid of process triggering event */
	uint32_t type;       /* event type ERTP_OPEN, ERTP_CLOSE or ERTP_EXEC, can be OR-ed with flags*/
	uint64_t file_id;    /* stat.st_ino - inode number */
	uint64_t device_number;  /* stat.st_dev - ID of device containing file */
	uint64_t fs_magic;   /* file system magic number from include/linux/magic.h, used for event's fs detection */
	uint64_t file_size;  /* file size in bytes */
	int64_t ctime;       /* tv_sec value of change time */
	int64_t mtime;       /* tv_sec value of modification time */
	uint16_t mode;       /* stat st_mode - type and mode */
};

/* copy eset_rtp_event */
static inline void eset_rtp_event_copy(struct eset_rtp_event *dst, struct eset_rtp_event *src)
{
	dst->size = src->size;
	dst->id = src->id;
	dst->fd = src->fd;
	dst->pid = src->pid;
	dst->type = src->type;
	dst->file_id = src->file_id;
	dst->device_number = src->device_number;
	dst->fs_magic = src->fs_magic;
	dst->file_size = src->file_size;
	dst->ctime = src->ctime;
	dst->mtime = src->mtime;
	dst->mode = src->mode;
}

struct eset_rtp_response {
	uint32_t size;   /* message size - for backward compatibility; more members can be added in future */
	int32_t  id;     /* id from eset_rtp_event */
	uint32_t result; /* ERTP_ALLOW or ERTP_DENY, can be OR-ed with ERTP_DO_NOT_CACHE */
	uint32_t scan_mask;  /* bit mask with applied scan settings */
};

#pragma pack()

/* How to use ertp
kernel module:
 *
 * 1. open(ERTP_DEV)
 * 2. watch ERTP_DEV with poll() for read
 * 3. pop struct eset_rtp_event with read from ERTP_DEV
 * 4. decide whether the file is clean or infected
 * 5. send struct eset_rtp_response with write on ERTP_DEV
 * 6. pop next eset_rtp_event, etc.
 *
 */

/* Paths to sysfs files */
#define ERTP_SETTINGS       "/sys/module/" ESET_RTP "/settings"
#define ERTP_EXCLUDES       ERTP_SETTINGS "/excludes"
#define ERTP_EXCLUDED_PROCS ERTP_EXCLUDES "/procs"
#define ERTP_EXCLUDED_FILES ERTP_EXCLUDES "/files"
#define ERTP_CACHE          ERTP_SETTINGS "/cache"
#define ERTP_CACHE_CONTROL  ERTP_CACHE "/control"
#define ERTP_SCAN_MASKS     ERTP_SETTINGS "/scan_masks"
#define ERTP_OPEN_EVENT_SCAN_MASK    ERTP_SCAN_MASKS "/open"
#define ERTP_CLOSE_EVENT_SCAN_MASK   ERTP_SCAN_MASKS "/close"
#define ERTP_EXEC_EVENT_SCAN_MASK    ERTP_SCAN_MASKS "/exec"

/* How to manage excludes:
 *
 * A) process excludes - ERTP_EXCLUDED_PROCS
 *  - read file to show excluded processes
 *  - write "a:<path>" to exclude processes executed from <paths>
 *  - write "c" to clean all proc excludes
 *
 * B) files/folders excludes - ERTP_EXCLUDED_FILES
 *  - read file to show excluded files/folders
 *  - write "a:<path>" to exclude files/folders from scanning
 *  - write "c" to clean all file excludes
 */

/* How to manage cache:
 *
 * All eset_rtp_response are automatically cached and used until file is changed.
 *
 * A) do not cache scan result
 *  - if eset_rtp_response should not be cached,
 *    please use ERTP_DO_NOT_CACHE flag in result.
 *
 * B) reset cache, e.g. when virus database has been updated
 *  - write "reset" into ERTP_CACHE_CONTROL
 */

#endif /* _ESET_RTP_H */
